# Copyright (C) 2007 Richard Musiol
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

require 'inline_acceleration'
require 'sexp'

class ParseTree
  inline do |builder|
    builder.include('"ruby.h"')
    builder.include('"st.h"')
    builder.include('"node.h"')
    builder.include('"env.h"')
    
    builder.prefix <<-CODE_END
      #define nd_3rd  u3.node
      #define nd_u1id u1.id
      #define nd_u2id u2.id
      
      struct BLOCK {
        NODE *var;
        NODE *body;
        VALUE self;
        struct FRAME frame;
        struct SCOPE *scope;
        VALUE klass;
        NODE *cref;
        int iter;
        int vmode;
        int flags;
        int uniq;
        struct RVarmap *dyna_vars;
        VALUE orig_thread;
        VALUE wrapper;
        VALUE block_obj;
        struct BLOCK *outer;
        struct BLOCK *prev;
      };
    CODE_END
    
    builder.link_function('NODE *',           'rb_compile_string',  ['char *', 'VALUE', 'int'])
    builder.link_function('ID',               'rb_to_id',           ['VALUE'])
    builder.link_function('struct RClass *',  'RCLASS',             ['VALUE'])
    builder.link_function('NODE *',           'RNODE',              ['long'])
    builder.link_function('int',              'st_lookup',          ['st_table *', 'st_data_t', 'st_data_t *']);
    builder.link_function('int',              'nd_type',            ['NODE *']);
    builder.link_function('int',              'nd_line',            ['NODE *']);
    builder.link_function('VALUE',            'ID2SYM',             ['ID']);
    builder.link_function('NODE *',           'NEW_SELF',           []);
    builder.link_function('void *',           'DATA_PTR',           ['VALUE']);
        
    builder.link_fields('struct RClass', {'m_tbl' => 'st_table *'})
    builder.link_fields('struct BLOCK',  {'var' => 'NODE *', 'body' => 'NODE *'})
    builder.link_fields('NODE',          {'nd_file' => 'char *', 'nd_mid' => 'ID', 'nd_vid' => 'ID', 'nd_u1id' => 'ID', 'nd_u2id' => 'ID', 'nd_lit' => 'VALUE', 'nd_cnt' => 'long', 'nd_rest' => 'long', 'nd_nth' => 'long', 'nd_tbl' => 'ID *', 'nd_body' => 'NODE *', 'nd_next' => 'NODE *', 'nd_head' => 'NODE *', 'nd_cpath' => 'NODE *', 'nd_super' => 'NODE *', 'nd_defn' => 'NODE *', 'nd_recv' => 'NODE *', 'nd_value' => 'NODE *', 'nd_args' => 'NODE *', 'nd_cond' => 'NODE *', 'nd_else' => 'NODE *', 'nd_iter' => 'NODE *', 'nd_var' => 'NODE *', 'nd_opt' => 'NODE *', 'nd_stts' => 'NODE *', 'nd_1st' => 'NODE *', 'nd_2nd' => 'NODE *', 'nd_3rd' => 'NODE *', 'nd_cval' => 'VALUE'})
    
    builder.link_constants('int', ['NODE_METHOD', 'NODE_FBODY', 'NODE_CFUNC', 'NODE_SCOPE', 'NODE_BLOCK', 'NODE_IF', 'NODE_CASE', 'NODE_WHEN', 'NODE_OPT_N', 'NODE_WHILE', 'NODE_UNTIL', 'NODE_ITER', 'NODE_FOR', 'NODE_BREAK', 'NODE_NEXT', 'NODE_REDO', 'NODE_RETRY', 'NODE_BEGIN', 'NODE_RESCUE', 'NODE_RESBODY', 'NODE_ENSURE', 'NODE_AND', 'NODE_OR', 'NODE_NOT', 'NODE_MASGN', 'NODE_LASGN', 'NODE_DASGN', 'NODE_DASGN_CURR', 'NODE_GASGN', 'NODE_IASGN', 'NODE_CDECL', 'NODE_CVASGN', 'NODE_CVDECL', 'NODE_OP_ASGN1', 'NODE_OP_ASGN2', 'NODE_OP_ASGN_AND', 'NODE_OP_ASGN_OR', 'NODE_CALL', 'NODE_FCALL', 'NODE_VCALL', 'NODE_SUPER', 'NODE_ZSUPER', 'NODE_ARRAY', 'NODE_ZARRAY', 'NODE_HASH', 'NODE_RETURN', 'NODE_YIELD', 'NODE_LVAR', 'NODE_DVAR', 'NODE_GVAR', 'NODE_IVAR', 'NODE_CONST', 'NODE_CVAR', 'NODE_NTH_REF', 'NODE_BACK_REF', 'NODE_MATCH', 'NODE_MATCH2', 'NODE_MATCH3', 'NODE_LIT', 'NODE_STR', 'NODE_DSTR', 'NODE_XSTR', 'NODE_DXSTR', 'NODE_EVSTR', 'NODE_DREGX', 'NODE_DREGX_ONCE', 'NODE_ARGS', 'NODE_ARGSCAT', 'NODE_ARGSPUSH', 'NODE_SPLAT', 'NODE_TO_ARY', 'NODE_SVALUE', 'NODE_BLOCK_ARG', 'NODE_BLOCK_PASS', 'NODE_DEFN', 'NODE_DEFS', 'NODE_ALIAS', 'NODE_VALIAS', 'NODE_UNDEF', 'NODE_CLASS', 'NODE_MODULE', 'NODE_SCLASS', 'NODE_COLON2', 'NODE_COLON3', 'NODE_CREF', 'NODE_DOT2', 'NODE_DOT3', 'NODE_FLIP2', 'NODE_FLIP3', 'NODE_ATTRSET', 'NODE_SELF', 'NODE_NIL', 'NODE_TRUE', 'NODE_FALSE', 'NODE_DEFINED', 'NODE_NEWLINE', 'NODE_POSTEXE', 'NODE_ALLOCA', 'NODE_DMETHOD', 'NODE_BMETHOD', 'NODE_MEMO', 'NODE_IFUNC', 'NODE_DSYM', 'NODE_ATTRASGN', 'NODE_LAST'])
  end
  
  node_symbols = []
  constants.each do |name|
    next if name[0..4] != 'NODE_'
    node_symbols[const_get(name)] = name.to_s[5..-1].downcase.to_sym
  end
  NODE_SYMBOLS = node_symbols
  
  @include_newlines = false;
  @level = 0
 
  def initialize(include_newlines = false)
    @include_newlines = include_newlines
  end
  
  def parse_tree_for_method(klass, method, is_cls_method = false, newlines = @include_newlines)
    @newlines = newlines
    @locals = []
    @current_sexp = nil
    method = method.to_sym
    id = rb_to_id(method)
    table = RCLASS(klass).m_tbl
    if (r, n = st_lookup(table, id))[0]
      sexp = [:defn, method, to_sexp(n.dereference.to_NODE.nd_body)]
      sexp.extend Sexp
      sexp
    else
      [nil]
    end
  end
  
  def parse_tree_for_string(string, filename = nil, line = nil, newlines = @include_newlines)
    filename ||= '(string)'
    line ||= 1
    @newlines = newlines
    @locals = []
    @current_sexp = nil
    
    node = rb_compile_string(filename, string, line)
    [to_sexp(node)]
  end
  
  def to_sexp(node)
    return nil if node.nil?
    c_type = nd_type(node)
    type = NODE_SYMBOLS[c_type]
    
    if type == :newline
      if @newlines
        newline_sexp = []
        newline_sexp << :newline      
        newline_sexp << nd_line(node)
        newline_sexp << node.nd_file
        @current_sexp << newline_sexp
      end
      return to_sexp(node.nd_next)
    end
    
    sexp = []
    sexp.extend Sexp
    last_current_sexp = @current_sexp
    @current_sexp = sexp
    sexp << type
    
    #    puts(('  ' * @level) + type.to_s)
    #    @level += 1
    
    case type
    when :block, :array
      begin
        sexp << to_sexp(node.nd_head)
      end while node = node.nd_next
    when :method
    when :fbody
    when :cfunc
    when :scope
      locals_before = @locals
      @locals = node.nd_tbl
      nd_next = node.nd_next
      sexp << to_sexp(nd_next) unless nd_next.nil?
      @locals = locals_before
    when :if
      sexp << to_sexp(node.nd_cond)
      sexp << to_sexp(node.nd_body)
      sexp << to_sexp(node.nd_else)
    when :case
      sexp << to_sexp(node.nd_head)
      child = node.nd_body
      loop do
        child_sexp = to_sexp(child)
        sexp << child_sexp
        if child_sexp[0] == :when #when clause
          child = child.nd_next
          if child.nil? #no else clause used
            sexp << nil
            break
          end
        else #else clause
          break
        end
      end
    when :when
      #    if (!case_level) { /* when without case, ie, no expr in case */
      #      rb_ary_pop(ary); /* reset what current is pointing at */
      #      node = NEW_CASE(0, node);
      #      goto again;
      #    }
      sexp << to_sexp(node.nd_head)
      sexp << to_sexp(node.nd_body)
    when :while, :until
      sexp << to_sexp(node.nd_cond)
      sexp << to_sexp(node.nd_body)
      sexp << !node.nd_3rd.nil?
    when :iter, :for
      sexp << to_sexp(node.nd_iter)
      #            if (node->nd_var != (NODE *)1
      #        && node->nd_var != (NODE *)2
      #        && node->nd_var != NULL) {
      if node.nd_var
        sexp << to_sexp(node.nd_var)
      else
        sexp << nil
      end
      sexp << to_sexp(node.nd_body)
    when :break, :next, :return, :yield
      stts = node.nd_stts
      sexp << to_sexp(stts) unless stts.nil?
    when :begin, :opt_n, :not
      sexp << to_sexp(node.nd_body)
    when :rescue
      sexp << to_sexp(node.nd_1st)
      sexp << to_sexp(node.nd_2nd)
      nd_3rd = node.nd_3rd
      sexp << to_sexp(nd_3rd) unless nd_3rd.nil?
    when :resbody
      sexp << to_sexp(node.nd_3rd)
      nd_2nd = node.nd_2nd
      sexp << to_sexp(nd_2nd) unless nd_2nd.nil?
      nd_1st = node.nd_1st
      sexp << to_sexp(nd_1st) unless nd_1st.nil?
    when :ensure
      sexp << to_sexp(node.nd_head)
      ensr = node.nd_ensr
      sexp << zo_sexp(ensr) unless ensr.nil?
    when :and, :or
      sexp << to_sexp(node.nd_1st)
      sexp << to_sexp(node.nd_2nd)
    when :masgn
      sexp << to_sexp(node.nd_head)
      args = node.nd_args
      sexp << to_sexp(args) unless args.nil?
      value = node.nd_value
      sexp << to_sexp(node.nd_value) unless value.nil?
    when :lasgn, :dasgn, :dasgn_curr, :gasgn, :iasgn, :cdecl, :cvasgn, :cvdecl
      sexp << ID2SYM(node.nd_vid)
      value = node.nd_value
      sexp << to_sexp(value) unless value.nil?
    when :op_asgn1
    when :op_asgn2
    when :op_asgn_and, :op_asgn_or
      sexp << to_sexp(node.nd_head)
      sexp << to_sexp(node.nd_value)
    when :call, :fcall, :vcall
      sexp << to_sexp(node.nd_recv) if type == :call
      sexp << ID2SYM(node.nd_mid)
      if type == :call || type == :fcall
        args = node.nd_args
        sexp << to_sexp(args) if args
      end
    when :super
      
    when :hash
      list = node.nd_head
      while list
        sexp << to_sexp(list.nd_head)
        list = list.nd_next
        raise "ParseTree error: Odd number list for Hash" if list.nil?
        sexp << to_sexp(list.nd_head)
        list = list.nd_next
      end
    when :lvar, :dvar, :gvar, :ivar, :const, :cvar, :attrset
      sexp << ID2SYM(node.nd_vid)
    when :nth_ref
      sexp << node.nd_nth
    when :back_ref
    when :match
    when :match2, :match3
      sexp << to_sexp(node.nd_recv)
      sexp << to_sexp(node.nd_value)
    when :lit, :str, :xstr
      sexp << node.nd_lit
    when :dstr, :dxstr, :dregx, :dregx_once, :dsym
      sexp << node.nd_lit
      list = node.nd_next
      while list
        head = list.nd_head
        if head
          child_sexp = to_sexp(head)
          child_sexp = child_sexp[1] if child_sexp[0] == :evstr
          sexp << child_sexp
        end
        list = list.nd_next
      end
    when :evstr
      sexp << to_sexp(node.nd_body)
    when :args
      arg_flag = node.nd_rest
      if @locals
        #regular args
        i = 0
        while i < node.nd_cnt do
          sexp << ID2SYM(@locals[i + 3])
          i += 1
        end
        #optional args
        optnode = node.nd_opt
        while optnode do
          sexp << ID2SYM(@locals[i + 3])
          i += 1
          optnode = optnode.nd_next
        end
        #special args
        case
        when arg_flag > 0: sexp << "*#{locals[i + 3].to_s}".to_sym #*arg
        when arg_flag == 0: #empty list
        when arg_flag == -1: #all handled above
        when arg_flag == -2: #no name
        else raise "ParseTree error: Unknown arg flag #{arg_flag}"
        end
        #assigns of optional args
        opt = node.nd_opt;
        if opt
          sexp << to_sexp(opt)
        end
      end
    when :argscat, :argspush
      sexp << to_sexp(node.nd_head)
      sexp << to_sexp(node.nd_body)
    when :splat, :to_ary, :svalue
      sexp << to_sexp(node.nd_head)
    when :block_arg
      sexp << ID2SYM(node.nd_u1id)
    when :block_pass
      sexp << to_sexp(node.nd_body)
      sexp << to_sexp(node.nd_iter)
    when :defn, :defs
      defn = node.nd_defn
      if defn
        sexp << to_sexp(node.nd_recv) if type == :defs
        sexp << ID2SYM(node.nd_mid)
        sexp << to_sexp(defn)
      end
    when :alias
    when :valias
    when :undef
    when :class, :module
      sexp << ID2SYM(node.nd_cpath.nd_mid)
      sexp << to_sexp(node.nd_super) if type == :class
      sexp << to_sexp(node.nd_body)
    when :sclass
    when :colon2
    when :colon3
    when :cref
    when :dot2
    when :dot3
    when :flip2
    when :flip3
    when :defined
    when :postexe
    when :alloca
    when :dmethod
    when :bmethod
      block = DATA_PTR(node.nd_cval).to_struct_BLOCK
      var = block.var
      sexp << to_sexp(var) unless var.nil?
      sexp << to_sexp(block.body)
    when :memo
    when :ifunc
    when :attrasgn
      nd_1st = node.nd_1st
      if nd_1st == RNODE(1)
        sexp << to_sexp(NEW_SELF())
      else
        sexp << to_sexp(nd_1st)
      end
      sexp << ID2SYM(node.nd_u2id)
      sexp << to_sexp(node.nd_3rd)
    when :last
    when :retry, :redo, :self, :nil, :true, :false, :zsuper, :zarray
      # empty nodes
    when :last
      raise "ParseTree error: Unsupported node type '#{type}'"
    else
      raise "ParseTree error: Unknown node type '#{c_type}'"
    end
    
    #    @level -= 1
    #    puts(('  ' * @level) + type.to_s + ' end')
    
    @current_sexp = last_current_sexp
    sexp
  end
end
